/* cfft.h - functions for reading Chilbolton FFT data */

#include <stdio.h>

#ifdef __cplusplus
extern "C" {
#endif                          /* __cplusplus */

#define CFFT_OK 0
#define CFFT_MEMORY_ERROR 1
#define CFFT_FILE_ERROR 2
#define CFFT_EOF 3

#define CFFT_NO_SIGNAL -1.0e10

#define CFFT_DEFAULTS 0
#define CFFT_USE_LOG_SNR (1L<<0) /* Convert to signal-to-noise in dB */
#define CFFT_REMOVE_NOISE (1L<<1) /* Set all noise to CFFT_NO_SIGNAL */
#define CFFT_REMOVE_SPIKE (1L<<2) /* Remove interference spike @ 0 m/s */
#define CFFT_REVERSE_V (1L<<3)

extern int cfft_error;
extern char cfft_verbose;

#ifdef CFFT_USE_FLOATS
typedef float CFFTFloat;
#else
typedef double CFFTFloat;
#endif

/*** STRUCTURES ***/
typedef struct {
  int ngates;
  int nbins;
  int naverages;
  int divfactor;
  int delayclocks;
  float offset1;
  float offset2;
  int interval;
} __CFFTOldHeader;

typedef struct {
  int file_header_fields;
  int profile_header_fields;
  int interval; /* ray interval of fft */
  int ngates;
  int nbins;
  int naverages;
  int divfactor;
  int delayclocks;
  float offset1;
  float offset2;
  float frequency_scaling;
} __CFFTHeader;

typedef struct {
  unsigned long _nallocated;
  unsigned long flags;
  CFFTFloat *spectra;
  double *rawspectra; /* only used if `CFFTFloat' is not `double' */
  CFFTFloat *range;
  CFFTFloat *velocity;
  CFFTFloat range_offset; /* In metres */
  CFFTFloat mean_noise; /* Linear units, from last call to cfft_process_spectra */
  CFFTFloat std_noise; /* Linear units, from last call to cfft_process_spectra */
  FILE *file;
  int ngates, nbins, naverages;
  int divfactor, delayclocks;
  float offset1, offset2; /* Channels 1 and 2 */
  int interval; /* between rays */
  int profile; /* Index of current profile (-1 = not started yet) */
  int noisegates;
  float noisefactor;
  unsigned int spikebleed;
  float starttime; /* In decimal hours, read from the .raw file */
  float rayduration; /* In hours, read from the .raw file */
  int file_header_size;
  int profile_header_size;
  unsigned short int startday, startmonth, startyear;
  unsigned short int hour, minute, second, day, month, year;
  char oldformat;
  unsigned char sizeof_real;
} CFFTData;

/*** PROTOTYPES ***/

#define cfft_verbose() cfft_verbose = 1

/* File operations */
int cfft_open(char *filename, CFFTData *data);
int cfft_close(CFFTData *data);

/* Reading data */
int cfft_next_profile(CFFTData *data);
int cfft_read_profile(CFFTData *data, int profile);

/* Reading raw file times */
int cfft_get_time(CFFTData *data, char *filename);
int cfft_sync_time(CFFTData *data);

/* Processing the spectra */
#define cfft_set_noise_processing(data, n, f) \
           (data)->noisegates = (n); (data)->noisefactor = (f)
#define cfft_set_processing_flags(data, f) \
           (data)->flags = (f)
#define cfft_set_range_offset(data, offs) (data)->range_offset = offs;
int cfft_process_spectra(CFFTData *data);

#ifdef __cplusplus
}                               /* extern "C" */
#endif                          /* __cplusplus */
