#include <fstream>
#include <cstring>
#include <cstdio>
#include <cmath>
#include <csignal>
#include "rplot.h"
extern "C" {
#include "chil.h"
}

#define ERROR(msg,status) cerr << "Error: " << (msg) << "\n"; exit(status)
#define FLUSH(msg) cerr << (msg) << ": flushing figure\n"; flush_and_exit()
#define WARNING(msg) cerr << "Warning: " << (msg) << "\n"
//#define DEBUG_LINE /* */
#define DEBUG_LINE CHILPrintf("DEBUG: %s got to line %d\n",__FILE__,__LINE__);
#define NO_SIGNAL -999.00
#ifndef X_FONT_SERVER
#define X_FONT_SERVER "unix/:7100"
#endif

enum FILE_FORMAT {
  FORMAT_POSTSCRIPT,
  FORMAT_PNG
};

RFigure *figure;
RAxes *axes, *cbar;
RPalette colormap;
RFontLoader *font_loader;

FILE_FORMAT format=FORMAT_PNG;
ostream *os=&cout;
char datestr[32], namestr[128], sysstr[16];

FileList filelist;
ChilboltonFile chil_file;
RasterInfo raster_info;
Ray ray;
float correct_range[1024];
float decimalhour, lastdecimalhour=-1, duration=6.0;
double starthour=0, endhour=24, minheight=0, maxheight=12,dhour=1.0;
double minZ=-40, maxZ=10;
bool verbose=false;
int daymod=0;

int Usage(char **argv) {
  cerr << "Usage: " << argv[0] << " [options] file001.raw file002.raw ...
Image is sent to stdout. Options:	
	-date		YYYYMMDD	only plot date from specified date
	-hourrange	hr1 hr2		set hour axis range
	-duration	hrs		set duration of plot (overridden by `-hourrange')
	-heightrange	ht1 ht2		set height axisrange
	-Zrange		Z1 Z2		set height axisrange
	-averaging	<rays>x<gates>	set averaging
	-minelev	el		set minimum elevation in degrees (default 80)
	-cal		cal		set Z calibration figure
	-noisefactor	f		set noise rejection threshold (default 3)
	-oscillator			correct for oscillator interference in Galileo
	-size		<x>x<y>		set size of figure
	-ps, -eps			set output format to encapsulated postscript
	-png				set output format to portable network graphics (default)
	-dots				send a `.' to stderr for each ray processed 
	-verbose			talk to me (also implements `-dots')
";
  exit(0);
}

void flush_and_exit() {

  CHILClose(&chil_file);
  CHILFreeRay(&ray);

  axes->box();
  axes->xtick(24,NULL,0,0,1);
  dhour = (endhour-starthour > 6) ? floor((endhour-starthour)/6.0) :
    (1.0+floor((endhour-starthour)/4.0))/2.0;
  axes->xticks_time(0,dhour);
  axes->yticks(0,1);
  axes->xlabel("Time (UTC)");
  axes->ylabel("Height (km)");

  if (format == FORMAT_PNG) {
    figure->write_png(&cout);
  }

  /* Must delete these objects in the right order for some unknown reason -
     oh, the joy of C++ */
  delete font_loader;
  delete axes;
  delete cbar;
  delete figure;

  exit(0);
}

void flush_on_signal(int signal) {
  FLUSH("Interupt received");
}


int main(int argc,char **argv) {
  unsigned short int reqday=0,reqmonth=0,reqyear=0,
    iarg=0;
  int i, iZ=-1, ifile, daymod=0;
  bool oscillator=false;
  bool bydate=false, sethourrange=false, setheightrange=false;
  bool plotdate=true, firstray=true, dots=false;
  int width=800,height=300;
  int averays=30, avegates=1, status;
  float Zcalibration=+25,noisefactor=3.0,minelev=85.0;

  /* Process command-line arguments */

  if (argc < 2) {
    Usage(argv);
  }

  while (++iarg < argc && *argv[iarg] == '-') {
    if (strcmp(argv[iarg],"-date") == 0) {
      bydate=true;
      if (sscanf(argv[++iarg],"%04hu%02hu%02hu",
		 &reqyear,&reqmonth,&reqday) < 3) {
	ERROR("Failed to read date: should be in the form YYYYMMDD",1);
      }
    }
    else if (strcmp(argv[iarg],"-hourrange") == 0) {
      if (sscanf(argv[++iarg],"%lf",&starthour) < 1) {
	ERROR("Failed to read start hour as double",1);
      }
      if (sscanf(argv[++iarg],"%lf",&endhour) < 1) {
	ERROR("Failed to read start hour as double",1);
      }
      sethourrange=true;
    }
    else if (strcmp(argv[iarg],"-heightrange") == 0) {
      if (sscanf(argv[++iarg],"%lf",&minheight) < 1) {
	ERROR("Failed to read minimum height as double",1);
      }
      if (sscanf(argv[++iarg],"%lf",&maxheight) < 1) {
	ERROR("Failed to read maximum height as double",1);
      }
      setheightrange=true;
    }
    else if (strcmp(argv[iarg],"-Zrange") == 0) {
      if (sscanf(argv[++iarg],"%lf",&minZ) < 1) {
	ERROR("Failed to read minimum Z as double",1);
      }
      if (sscanf(argv[++iarg],"%lf",&maxZ) < 1) {
	ERROR("Failed to read maximum Z as double",1);
      }
      if (minZ >= maxZ) {
	ERROR("Minimum Z must be less than maximum Z",1);
      }
    }
    else if (strcmp(argv[iarg],"-size") == 0) {
      if (sscanf(argv[++iarg],"%dx%d",&width,&height) < 2) {
	ERROR("Failed to read size: should be in the form <width>x<height>",1);
      }
    }
    else if (strcmp(argv[iarg],"-averaging") == 0) {
      if (sscanf(argv[++iarg],"%dx%d",&averays,&avegates) < 2) {
	ERROR("Failed to read averaging: should be in the form <rays>x<gates>",1);
      }
      if (averays < 1) {
	WARNING("Cannot average negative or zero rays");
	averays = 1;
      }
      if (avegates < 1) {
	WARNING("Cannot average negative or zero gates");
	avegates = 1;
      }
    }
    else if ((strcmp(argv[iarg],"-ps") == 0) || (strcmp(argv[iarg],"-eps") == 0)) {
      format = FORMAT_POSTSCRIPT;
    }
    else if (strcmp(argv[iarg],"-png") == 0) {
      format = FORMAT_PNG;
    }
    else if (strcmp(argv[iarg],"-oscillator") == 0) {
      oscillator=true;
    }
    else if (strcmp(argv[iarg],"-duration") == 0) {
      duration=atof(argv[++iarg]);
    }
    else if (strcmp(argv[iarg],"-cal") == 0 && iarg<argc-1) {
      Zcalibration=atof(argv[++iarg]);
    }
    else if (strcmp(argv[iarg],"-noisefactor") == 0) {
      noisefactor=atof(argv[++iarg]);
    }
    else if (strcmp(argv[iarg],"-minelev") == 0) {
      minelev=atof(argv[++iarg]);
    }
    else if (strcmp(argv[iarg],"-verbose") == 0) {
      verbose=true;
    }
    else if (strcmp(argv[iarg],"-dots") == 0) {
      dots=true;
    }
    else {
      cerr << "Error: `" << argv[iarg] << "' not understood\n";
      Usage(argv);
    }
  }

  if (argc < iarg) {
    Usage(argv);
  }

  if (bydate) {
    sethourrange=true;
  }

  try {

    CHILSetNoSignalValue(NO_SIGNAL);

    if (format == FORMAT_POSTSCRIPT) {
      figure = new RFigure(width,height,POSTSCRIPT,os,
			   "Chilbolton cloud radar");
    }
    else {
      figure = new RFigure(width,height,BITMAP_8BIT);
    }

    font_loader = new RFontLoader(X_FONT_SERVER, figure->is_bitmap());

    font_loader->verbose(verbose);

    figure->clear();
    figure->axis(0,0,1,1);
    figure->font = font_loader->load(12,string("helvetica"),string("medium"),
				     string("r"),string("adobe"));

    create_jet_palette(&colormap,64);

    cbar = new RAxes(figure,0.96,0.23,0.98,0.8);
    cbar->colorbar(&colormap,VERTICAL);
    cbar->box();
    cbar->axis(0,minZ,1,maxZ);
    cbar->yticks(minZ,10,"%g",0);
    cbar->ylabel("Radar reflectivity factor (dBZ)");

    axes = new RAxes(figure,0.07,0.13,0.89,0.9);

    signal(SIGINT,flush_on_signal);
    signal(SIGTERM,flush_on_signal);

    for (ifile=iarg;ifile<argc;ifile++) {
      if (CHILOpen(argv[ifile],&chil_file) == FILE_CLOSED) {
	FLUSH("Error reading file");
      }
      if (verbose) {
	CHILPrintFileInformation(&chil_file);
      }

      while (~feof(chil_file.file) && 
	     CHILReadInformationRecord(&chil_file,&raster_info)
	     != END_OF_DATA) {

	if (verbose) {
	  CHILPrintRasterInformation(&raster_info);
	}
	if (firstray) {
	  firstray=false;
	  CHILInitRay(&raster_info,avegates,averays,&ray);
	  if (verbose) {
	    CHILPrintRayInformation(&ray);
	  }
	  /* Find which parameter is Z */
	  iZ=-1;
	  for (i=0; i<ray.nfields; i++) {
	    if (ray.field_type[i] == ZH) {
	      iZ=i;
	      break;
	    }
	  }
	  if (iZ < 0) {
	    FLUSH("Z not found in file");
	  }
	  if (!sethourrange) {
	    starthour=raster_info.hour 
	      + (((float) raster_info.minute)/60.0) 
	      + (((float) raster_info.second)/3600.0);
	    endhour=starthour+duration;
	  }
	  if (endhour <= starthour) {
	    FLUSH("Error: End hour is less than start hour!");
	  }

	  CHILSystemName(&raster_info, sysstr);
	  sprintf(namestr,"<B>Chilbolton %g GHz Cloud Radar (%s)",
		  raster_info.frequency, sysstr);
	  figure->text(0.01,0.0,namestr,
		       ANCHOR_BOTTOMLEFT);

	  axes->axis(starthour,minheight,endhour,maxheight);
	  axes->clipbox_to_axis();
	  //			CHILGetDate(datestr,&ray);
	  //			axes->text(starthour,maxheight,datestr,ANCHOR_BOTTOMLEFT);
	}
	else {
	  CHILReInitRay(&raster_info,avegates,averays,&ray);
	}

	for (i=0;i<ray.ngates;i++) {
	  correct_range[i] = ray.range[i] + ray.range_offset*0.001;
	}
	if (verbose || dots) {
	  cerr << "Raster " << (int) raster_info.raster;
	}

	while ((status=CHILReadRay(&chil_file,&raster_info,&ray)) == CHIL_OK
	       && ~feof(chil_file.file)
	       ) {

	  /* fprintf(stderr,"%d %d %d %d %d %d\n",ray.day, ray.month, CHILTrueYear(ray.year), reqday,
	     reqmonth, reqyear); */
	  if ( (!bydate || (ray.day == reqday && ray.month == reqmonth &&
			    CHILTrueYear(ray.year) == reqyear )) 
	       && ray.elevation >= minelev) {
	    decimalhour = ray.utc+ 24*daymod;
	    plotdate=(lastdecimalhour<0);
	    if (lastdecimalhour>decimalhour) {
	      daymod++;
	      decimalhour += 24;
	      plotdate=true;
	    }
	    lastdecimalhour=decimalhour;
	    if ((decimalhour>endhour) && (!firstray)) {
	      break;
	    }
	    else if (decimalhour<starthour) {
	      continue;
	    }
	    if (plotdate) {
	      CHILGetDate(datestr,&ray);
	      axes->text(decimalhour,maxheight,datestr,
			 ANCHOR_BOTTOMLEFT);
	    }
	    if (oscillator) {
	      if (CHILSubtractOscillatorNoise(&ray,25,noisefactor) != 0 ) {
		FLUSH("Cannot subtract oscillator noise");
	      }
	    }
	    else {
	      if (CHILSubtractNoise(&ray,25,noisefactor)
		  != 0 ) {
		FLUSH("Cannot subtract noise");
	      }
	    }
	    CHILCleanRay(&ray);
	    CHILRangeCorrection(&raster_info,&ray);
				/* Calibrate Z... */

	    for (i=0;i<ray.ngates;i++) {
	      if ( *(ray.data[iZ]+i) != NO_SIGNAL) {
		(*(ray.data[iZ]+i)) += Zcalibration;
	      }
	    }
	    if (verbose || dots) {
	      cerr << ".";
	    }
	    axes->vray(decimalhour, decimalhour+averays/1800.0,
		       correct_range, ray.data[iZ], ray.ngates, &colormap, minZ, 10,
		       NO_SIGNAL);
	  }
	}
	if (verbose || dots) {
	  cerr << "\n";
	}

	if (chil_file.is_raw) {
	  break;
	}

      }
      CHILClose(&chil_file);
    }

    flush_and_exit();
  }
  catch  (char * error_message) {
    cerr << error_message << endl;
    exit(1);
  }

  exit(0);
}




